@@ -11,6 +11,15 @@
 
 #include "defs.h"
 
+/*
+ * Try to turn back the IPv6 monster at the gate
+ */
+#ifdef DO_NOT_USE_IPV6
+# undef INET6
+#else
+# define INET6
+#endif
+
 /*
  * Everybody needs these ANSI headers...
  */
@@ -19,6 +28,12 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include <stdarg.h>
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
+#ifdef HAVE_INTTYPES_H
+#include <inttypes.h>
+#endif
 
 /*
  * Everybody needs these POSIX headers...
@@ -29,6 +44,7 @@
 #include <limits.h>
 #include <sys/param.h>
 #include <errno.h>
+#include <sys/stat.h>
 
 /*
  * Everybody needs these INET headers...
@@ -123,7 +139,9 @@
  * I took this from the autoconf documentation
  */
 #if defined(__GNUC__) && !defined(HAVE_ALLOCA_H)
-# define alloca __builtin_alloca
+# ifndef alloca
+#  define alloca __builtin_alloca
+# endif
 #else
 # if HAVE_ALLOCA_H
 #  include <alloca.h>
@@ -159,6 +177,13 @@ extern	char	*sys_errlist[];
 #define strerror(x) sys_errlist[x]
 #endif
 
+/*
+ * Deal with brokenness with realpath.
+ */
+#ifdef HAVE_BROKEN_REALPATH
+# define realpath my_realpath
+#endif
+
 /*
  * Dont trust anyone else's NULLs.
  */
@@ -167,6 +192,13 @@ extern	char	*sys_errlist[];
 #endif
 #define NULL (void *) 0
 
+/*
+ * Make sure there is TRUE and FALSE
+ */
+#ifndef TRUE
+#define TRUE 1
+#define FALSE 0
+#endif
 
 /*
  * Can you believe some systems done #define this?
@@ -193,7 +225,15 @@ sigfunc *my_signal (int, sigfunc *);
 #define BUILT_IN_COMMAND(x) \
 	void x (const char *command, char *args, const char *subargs)
 
+typedef char Filename[MAXPATHLEN + 1];
 
+/*
+ * It's really really important that you never use LOCAL_COPY in the actual
+ * argument list of a function call, because bad things can happen.  Always
+ * do your LOCAL_COPY as a separate step before you call a function.
+ */
+#define LOCAL_COPY(y) strcpy(alloca(strlen((y)) + 1), y)
+#define SAFE(x) (((x) && *(x)) ? (x) : empty_string)
 
 /*
  * Deal with our brokenness wrt ANSI.  Sigh.
@@ -250,5 +290,74 @@ sigfunc *my_signal (int, sigfunc *);
 #include <sys/sysctl.h>
 #endif
 
+/*
+ * Define an RFC2553 compatable "struct sockaddr_storage" if we do not
+ * already have one.
+ */
+#ifndef HAVE_STRUCT_SOCKADDR_STORAGE
+struct sockaddr_storage {
+#ifdef HAVE_SA_LEN
+	u_char ss_len;
+	u_char ss_family;
+#else
+	u_short ss_family;
+#endif
+	u_char padding[128 - 2];
+};
+#endif
+
+#ifndef HAVE_SOCKLEN_T
+typedef int socklen_t;
+#endif
+
+#ifndef HAVE_STRUCT_SOCKADDR_IN6
+#undef INET6
+#endif
+
+#if !defined(HAVE_GETADDRINFO) || !defined(HAVE_GETNAMEINFO) || !defined(HAVE_STRUCT_ADDRINFO)
+# define NEED_GAILIB
+# undef INET6
+# ifndef HAVE_GETADDRINFO
+#  define getaddrinfo getaddrinfo__compat
+#  define freeaddrinfo freeaddrinfo__compat
+#  define gai_strerror gai_strerror__compat
+   struct addrinfo__compat {
+        int     ai_flags;       /* AI_PASSIVE, AI_CANONNAME */
+        int     ai_family;      /* PF_xxx */
+        int     ai_socktype;    /* SOCK_xxx */
+        int     ai_protocol;    /* 0 or IPPROTO_xxx for IPv4 and IPv6 */
+        size_t  ai_addrlen;     /* length of ai_addr */
+        char    *ai_canonname;  /* canonical name for hostname */
+        struct sockaddr *ai_addr;       /* binary address */
+        struct addrinfo__compat *ai_next;    /* next structure in linked list */
+   };
+#  define addrinfo addrinfo__compat
+# endif
+# ifndef HAVE_GETNAMEINFO
+#  define getnameinfo getnameinfo__compat
+# endif
+# include "gailib.h"
+#endif
+
+
+/*
+ * Define some lazy shorthand typedefs for commonly used structures
+ */
+typedef struct sockaddr 	SA;
+typedef struct sockaddr_storage	SS;
+typedef struct sockaddr_in 	ISA;
+typedef struct in_addr		IA;
+
+#ifdef INET6
+typedef struct sockaddr_in6	ISA6;
+typedef struct sockaddr_in6	I6SA;
+typedef struct in6_addr		IA6;
+typedef struct in6_addr		I6A;
+#endif
+
+typedef struct addrinfo		AI;
+typedef struct hostent		Hostent;
+typedef struct timeval		Timeval;
+typedef struct stat		Stat;
 
 #endif /* __irc_std_h */
