@@ -22,13 +22,19 @@ typedef struct notify_alist
 	char *			ison;
 } NotifyList;
 
+extern	char	notify_timeref[];
+
 	BUILT_IN_COMMAND(notify);
 	void	do_notify 		(void);
-	void	notify_mark 		(char *, int, int);
+	void	notify_mark 		(int, const char *, int, int);
 	void	save_notify 		(FILE *);
 	void	set_notify_handler 	(char *);
 	void	make_notify_list 	(int);
 	char *	get_notify_nicks 	(int, int);
 	void	destroy_notify_list	(int);
 
+	void	notify_systimer		(void);
+	void	set_notify_interval	(const void *);
+	void	set_notify		(const void *);
+
 #endif /* _NOTIFY_H_ */
