@@ -1,3 +1,4 @@
+/* $EPIC: compat.c,v 1.23 2004/02/20 22:54:16 jnelson Exp $ */
 /*
  * Everything that im not directly responsible for I put in here.  Almost
  * all of this stuff is either borrowed from somewhere else (for you poor
@@ -38,13 +39,16 @@
  * SUCH DAMAGE. 
  * * * *
  *
- * And then everthing is either explicitly placed in the public domain
- * by someone else, or placed into the public domain by me.
+ * The second license is OpenBSD's ISC-like license, which is used for 
+ * strlcpy() and strlcat().  See the license later on in the file.
+ * 
+ * Everthing else has had its copyright explicitly disclaimed by the author.
  */
 
 
 #include "defs.h"
 #include "irc_std.h"
+#include "ircaux.h"
 
 /*****************************************************************************/
 /* ------------------------------- start of tparm.c ------------------------ */
@@ -163,7 +167,7 @@ static int popnum(int *num)
 	return 1;
 }
 
-static int cvtchar(char *sp, char *c)
+static int cvtchar (const char *sp, char *c)
 {
 	switch(*sp) {
 	case '\\':
@@ -478,7 +482,7 @@ char *tparm(const char *str, ...) {
 				if (conv_char == 's') {
 					if (popstring(&s))
 						return OOPS;
-					sprintf(sbuf, fmt, s);
+					snprintf(sbuf, MAX_LINE, fmt, s);
 				} else {
 					if (termcap) {
 						if (getarg(termcap++ - 1,
@@ -490,7 +494,7 @@ char *tparm(const char *str, ...) {
 					if (i == 0 && conv_char == 'c')
 						*sbuf = 0;
 					else
-						sprintf(sbuf, fmt, i);
+						snprintf(sbuf, MAX_LINE, fmt, i);
 				}
 				sp++;
 				fmt = sbuf;
@@ -885,183 +889,68 @@ unsigned long strtoul (const char *nptr, char **endptr, int base)
 #endif /* DO NOT HAVE STRTOUL */
 /* ----------------------- end of strtoul.c ----------------------------- */
 /**************************************************************************/
-/* --------------------- start of inet_aton.c --------------------------- */
-#ifndef HAVE_INET_ATON
-/*
- * Copyright (c) 1983, 1990, 1993
- *    The Regents of the University of California.  All rights reserved.
- * Portions Copyright (c) 1993 by Digital Equipment Corporation.
- * Licensed under the 3-clause BSD license, see above for text.
- */
-
-/* 
- * Check whether "cp" is a valid ascii representation
- * of an Internet address and convert to a binary address.
- * Returns 1 if the address is valid, 0 if not.
- * This replaces inet_addr, the return value from which
- * cannot distinguish between failure and a local broadcast address.
- */
-int inet_aton(const char *cp, struct in_addr *addr)
-{
-	unsigned long	val;
-	int		base, n;
-	char		c;
-	unsigned	parts[4];
-	unsigned	*pp = parts;
-
-	c = *cp;
-	for (;;) {
-		/*
-		 * Collect number up to ``.''.
-		 * Values are specified as for C:
-		 * 0x=hex, 0=octal, isdigit=decimal.
-		 */
-		if (!isdigit(c))
-			return (0);
-		val = 0; base = 10;
-		if (c == '0') {
-			c = *++cp;
-			if (c == 'x' || c == 'X')
-				base = 16, c = *++cp;
-			else
-				base = 8;
-		}
-		for (;;) {
-			if (isascii(c) && isdigit(c)) {
-				val = (val * base) + (c - '0');
-				c = *++cp;
-			} else if (base == 16 && isascii(c) && isxdigit(c)) {
-				val = (val << 4) |
-					(c + 10 - (islower(c) ? 'a' : 'A'));
-				c = *++cp;
-			} else
-				break;
-		}
-		if (c == '.') {
-			/*
-			 * Internet format:
-			 *	a.b.c.d
-			 *	a.b.c	(with c treated as 16 bits)
-			 *	a.b	(with b treated as 24 bits)
-			 */
-			if (pp >= parts + 3)
-				return (0);
-			*pp++ = val;
-			c = *++cp;
-		} else
-			break;
-	}
-	/*
-	 * Check for trailing characters.
-	 */
-	if (c != '\0' && (!isascii(c) || !isspace(c)))
-		return (0);
-	/*
-	 * Concoct the address according to
-	 * the number of parts specified.
-	 */
-	n = pp - parts + 1;
-	switch (n) {
-
-	case 0:
-		return (0);		/* initial nondigit */
-
-	case 1:				/* a -- 32 bits */
-		break;
-
-	case 2:				/* a.b -- 8.24 bits */
-		if (val > 0xffffff)
-			return (0);
-		val |= parts[0] << 24;
-		break;
-
-	case 3:				/* a.b.c -- 8.8.16 bits */
-		if (val > 0xffff)
-			return (0);
-		val |= (parts[0] << 24) | (parts[1] << 16);
-		break;
-
-	case 4:				/* a.b.c.d -- 8.8.8.8 bits */
-		if (val > 0xff)
-			return (0);
-		val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
-		break;
-	}
-	if (addr)
-		addr->s_addr = htonl(val);
-	return (1);
-}
-#endif
-/* --------------------- end of inet_aton.c ---------------------------- */
-/**************************************************************************/
 /* ---------------------- strlcpy, strlcat ----------------------------- */
 /*
  * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
- * All rights reserved.
- * Licensed under the 3-clase BSD license
+ *
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND TODD C. MILLER DISCLAIMS ALL
+ * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
+ * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TODD C. MILLER BE LIABLE
+ * FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
+ * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
+ * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #ifndef HAVE_STRLCPY
+/*	OpenBSD's strlcpy.c version 1.7 2003/04/12 21:56:39 millert */
 /*
  * Copy src to string dst of size siz.  At most siz-1 characters
  * will be copied.  Always NUL terminates (unless siz == 0).
  * Returns strlen(src); if retval >= siz, truncation occurred.
  */
-size_t strlcpy (char *dst, const char *src, size_t siz)
+size_t	strlcpy (char *dst, const char *src, size_t siz)
 {
-        char *d = dst;
-        const char *s = src;
-        size_t n = siz;
+	char *d = dst;
+	const char *s = src;
+	size_t n = siz;
 
-	if (n > 0)
-		n--;		/* Save space for the NUL */
+	/* Copy as many bytes as will fit */
+	if (n != 0 && --n != 0) {
+		do {
+			if ((*d = *s) == 0)
+				break;
+			d++;
+			s++;
+		} while (--n != 0);
+	}
 
-        /* Copy as many bytes as will fit */
-	while (n > 0)
-	{
-		if (!(*d = *s))
-			break;
-		d++;
-		s++;
-		n--;
-        }
+	/* Not enough room in dst, add NUL and traverse rest of src */
+	if (n == 0) {
+		if (siz != 0)
+			*d = '\0';		/* NUL-terminate dst */
+		while (*s)
+			s++;
+	}
 
-        /* Not enough room in dst, add NUL and traverse rest of src */
-        if (n == 0 && siz)
-                *d = 0;              /* NUL-terminate dst */
-
-	return strlen(src);
-
-#ifdef notanymore
-        /* Copy as many bytes as will fit */
-        if (n != 0 && --n != 0) {
-                do {
-                        if ((*d++ = *s++) == 0)
-                                break;
-                } while (--n != 0);
-        }
-
-        /* Not enough room in dst, add NUL and traverse rest of src */
-        if (n == 0) {
-                if (siz != 0)
-                        *d = '\0';              /* NUL-terminate dst */
-                while (*s++)
-                        ;
-        }
-
-        return(s - src - 1);    /* count does not include NUL */
-#endif
+	return(s - src);	/* count does not include NUL */
 }
 #endif
 
 #ifndef HAVE_STRLCAT
+/*      OpenBSD's strlcat.c version 1.10 2003/04/12 21:56:39 millert */
 /*
  * Appends src to string dst of size siz (unlike strncat, siz is the
  * full size of dst, not space left).  At most siz-1 characters
- * will be copied.  Always NUL terminates (unless siz == 0).
- * Returns strlen(src); if retval >= siz, truncation occurred.
+ * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
+ * Returns strlen(src) + MIN(siz, strlen(initial dst)).
+ * If retval >= siz, truncation occurred.
  */
-size_t strlcat(char *dst, const char *src, size_t siz)
+size_t	strlcat (char *dst, const char *src, size_t siz)
 {
         char *d = dst;
         const char *s = src;
@@ -1085,10 +974,143 @@ size_t strlcat(char *dst, const char *src, size_t siz)
         }
         *d = '\0';
 
-        return(dlen + (s - src));       /* count does not include NUL */
+        return(dlen + (s - src));        /* count does not include NUL */
 }
 #endif
 
+/* --------------------------- start of arc4 stuff -------------------- */
+/*
+ * Arc4 random number generator for OpenBSD.
+ * Copyright 1996 David Mazieres <dm@lcs.mit.edu>.
+ *
+ * Modification and redistribution in source and binary forms is
+ * permitted provided that due credit is given to the author and the
+ * OpenBSD project (for instance by leaving this copyright notice
+ * intact).
+ */
+
+/*
+ * This code is derived from section 17.1 of Applied Cryptography,
+ * second edition, which describes a stream cipher allegedly
+ * compatible with RSA Labs "RC4" cipher (the actual description of
+ * which is a trade secret).  The same algorithm is used as a stream
+ * cipher called "arcfour" in Tatu Ylonen's ssh package.
+ *
+ * Here the stream cipher has been modified always to include the time
+ * when initializing the state.  That makes it impossible to
+ * regenerate the same random sequence twice, so this can't be used
+ * for encryption, but will generate good random numbers.
+ *
+ * RC4 is a registered trademark of RSA Laboratories.
+ */
+struct bsd_arc4_stream {
+	u_char	i;
+	u_char	j;
+	u_char	s[256];
+};
+typedef struct bsd_arc4_stream 	ARC4;
+
+static int	rs_initialized = 0;
+static ARC4	rs;
+
+__inline__
+static void	bsd_arc4_init (ARC4 *as)
+{
+	int     n;
+
+	for (n = 0; n < 256; n++)
+		as->s[n] = n;
+	as->i = 0;
+	as->j = 0;
+}
+
+__inline__
+static void	bsd_arc4_addrandom (ARC4 *as, u_char *dat, int datlen)
+{
+	int     n;
+	u_char	si;
+
+	as->i--;
+	for (n = 0; n < 256; n++) {
+		as->i = (as->i + 1);
+		si = as->s[as->i];
+		as->j = (as->j + si + dat[n % datlen]);
+		as->s[as->i] = as->s[as->j];
+		as->s[as->j] = si;
+	}
+}
+
+static void	bsd_arc4_stir (ARC4 *as)
+{
+	int     fd;
+	struct {
+		Timeval tv;
+		pid_t 	pid;
+		u_char	rnd[128 - sizeof(Timeval) - sizeof(pid_t)];
+	}       rdat;
+
+	gettimeofday(&rdat.tv, NULL);
+	rdat.pid = getpid();
+	if ((fd = open("/dev/urandom", O_RDONLY, 0)) >= 0) {
+		read(fd, rdat.rnd, sizeof(rdat.rnd));
+		close(fd);
+	}
+	/* 
+	 * fd < 0?  Ah, what the heck. We'll just take whatever was on the
+	 * stack... 
+	 */
+	bsd_arc4_addrandom(as, (void *) &rdat, sizeof(rdat));
+}
+
+__inline__
+static u_char		bsd_arc4_getbyte (ARC4 *as)
+{
+	u_char si, sj;
+
+	as->i = (as->i + 1);
+	si = as->s[as->i];
+	as->j = (as->j + si);
+	sj = as->s[as->j];
+	as->s[as->i] = sj;
+	as->s[as->j] = si;
+	return (as->s[(si + sj) & 0xff]);
+}
+
+__inline__
+static u_32int_t	bsd_arc4_getword (ARC4 *as)
+{
+	u_32int_t val;
+
+	val = bsd_arc4_getbyte(as) << 24;
+	val |= bsd_arc4_getbyte(as) << 16;
+	val |= bsd_arc4_getbyte(as) << 8;
+	val |= bsd_arc4_getbyte(as);
+	return val;
+}
+
+void	bsd_arc4random_stir (void)
+{
+	if (!rs_initialized) {
+		bsd_arc4_init(&rs);
+		rs_initialized = 1;
+	}
+	bsd_arc4_stir(&rs);
+}
+
+void	bsd_arc4random_addrandom (u_char *dat, int datlen)
+{
+	if (!rs_initialized)
+		bsd_arc4random_stir();
+	bsd_arc4_addrandom(&rs, dat, datlen);
+}
+
+u_32int_t	bsd_arc4random (void)
+{
+	if (!rs_initialized)
+		bsd_arc4random_stir();
+	return bsd_arc4_getword(&rs);
+}
+
 /* --------------------------- start of misc stuff -------------------- */
 /* This is all written by Jeremy Nelson and is public domain */
 #ifndef HAVE_VSNPRINTF
@@ -1097,11 +1119,14 @@ int vsnprintf (char *str, size_t size, const char *format, va_list ap)
 	int ret = vsprintf(str, format, ap);
 
 	/* If the string ended up overflowing, just give up. */
+	/* Pre-ansi vsprintf()s return (char *) */
 	if (ret == (int)str && strlen(str) > size)
 		panic("Buffer overflow in vsnprintf");
+	/* ANSI vsprintf()s return (int) */
 	if (ret != (int)str && ret > size)
 		panic("Buffer overflow in vsnprintf");
 
+	/* We always return (int). */
 	return ret;
 }
 #endif
@@ -1126,27 +1151,21 @@ int	setsid	(void)
 }
 #endif
 
-#ifndef HAVE_STPCPY
-char *	stpcpy (char *to, const char *from)
-{
-	while ((*to = *from))
-		to++, from++;
-	return to;
-}
-#endif
-
 #ifndef HAVE_SETENV
 int	setenv (const char *name, const char *value, int overwrite)
 {
 	static int warning = 0;
 	char *envvalue;
+	size_t	len;
 
 	if (warning == 0) {
 		yell("Warning: Your system does not have setenv(3).  Setting the same environment variable multiple times will result in memory leakage.  This is unavoidable and does not represent a bug in EPIC.");
 		warning = 1;
 	}
 
-	envvalue = m_sprintf("%s=%s", name, value);
+	len = strlen(name) + strlen(value) + 2;
+	envvalue = (char *)malloc(len);
+	snprintf(envvalue, len, "%s=%s", name, value);
 	putenv(envvalue);
 	return 0;
 }
@@ -1160,6 +1179,75 @@ int	unsetenv (const char *name)
 }
 #endif
 
+#ifdef HAVE_BROKEN_REALPATH
+# if defined(realpath)
+#  undef realpath
+# endif
+char *	my_realpath (const char *pathname, char resolved_path[MAXPATHLEN])
+{
+	char *mypath;
+	char *rest;
+	Stat unused;
+	size_t	size;
 
-/* --- end of misc stuff --- */
+	/* If the file exists, just run realpath on it. */
+	if (stat(pathname, &unused) == 0)
+		return realpath(pathname, resolved_path);
+
+	/* Otherwise, run realpath() on the dirname only */
+	size = strlen(pathname) + 1;
+	mypath = alloca(size);
+	strlcpy(mypath, pathname, size);
+	if ((rest = strrchr(mypath, '/')))
+		*rest++ = 0;
+	else
+	{
+		rest = LOCAL_COPY(mypath);
+		strlcpy(mypath, ".", size);
+	}
+
+	if (realpath(mypath, resolved_path) == NULL)
+		return NULL;
+
+	/* And put the basename back on the result. */
+	strlcat(resolved_path, "/", MAXPATHLEN);
+	strlcat(resolved_path, rest, MAXPATHLEN);
+	return resolved_path;
+}
+#endif
+
+#ifndef HAVE_ATOLL
+# ifdef HAVE_LONG_LONG
+#  ifdef HAVE_STRTOLL
+#   define HAVE_ATOLL_REPLACEMENT
+long long	atoll (const char *str)
+{
+	return strtoll(str, NULL, 0);
+}
+#  else
+#   ifdef HAVE_ATOQ
+#    define HAVE_ATOLL_REPLACEMENT
+long long	atoll (const char *str)
+{
+	return (long long)atoq(str);
+}
+#   endif
+#  endif
+# endif
+# ifndef HAVE_ATOLL_REPLACEMENT
+#  ifdef HAVE_LONG_LONG
+long long	atoll (const char *str)
+{
+	return (long long)atol(str);
+}
+#  else
+long	atoll (const char *str)
+{
+	return atol(str);
+}
+#  endif
+# endif
+#endif
+
+/**** END MISC PUBLIC DOMAIN STUFF ****/
 
